(* Author:  Christian Sternagel *)

text \<open>
  Summary of induction proofs from last lecture.
  This time in Isabelle/HOL.
\<close>

theory Summary
imports
  Main
  (*"~~/src/HOL/Library/Code_Binary_Nat"*)
  "~~/src/HOL/Library/Code_Target_Nat"
begin

declare [[names_short]]


section \<open>Examples\<close>

lemma example1:
  fixes n :: nat
  shows "\<Sum>{0..n} = n * (n + 1) div 2"
by (induct n) auto

datatype 'a list = Nil | Cons 'a "'a list"

fun append :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
where
  "append Nil ys = ys"
| "append (Cons x xs) ys = Cons x (append xs ys)"

lemma example2:
  "append xs Nil = xs"
by (induct xs) auto

fun length :: "'a list \<Rightarrow> nat"
where
  "length Nil = 0"
| "length (Cons x xs) = length xs + 1"

lemma example3:
  "length (append xs ys) = length xs + length ys"
by (induct xs) auto

datatype 'a btree = Empty |  Node "'a btree" 'a "'a btree"

fun height :: "'a btree \<Rightarrow> nat"
where
  "height Empty = 0"
| "height (Node l x r) = max (height l) (height r) + 1"

fun size :: "'a btree \<Rightarrow> nat"
where
  "size Empty = 0"
| "size (Node l x r) = size l + size r + 1"

fun perfect :: "'a btree \<Rightarrow> bool"
where
  "perfect Empty \<longleftrightarrow> True"
| "perfect (Node l x r) \<longleftrightarrow> height l = height r \<and> perfect l \<and> perfect r"

lemma example4:
  "perfect t \<longrightarrow> size t = (2 ^ height t) - 1"
by (induct t) (auto simp: diff_Suc split: nat.splits)


section \<open>Code Generation\<close>

export_code open
  append length height
in OCaml (*Haskell, SML (Standard ML), Scala*)
module_name Test

end
